/*The code was written and debugged by Matt Gray, Michael Koppelman,
  and John C. Martin at the University of Minnesota as part of the Eta
  Carinae HST Treasury Project (copyright 2006).  You are free to
  distribute this source as long as this notice remains intact.  End
  users of this code agree to take responsibility for how this code is
  used and the programmers, the Eta Carinae HST Treasury Project, and the
  University of Minnesota bear no liability.*/  

#include "exfits.h"

extern FILE *logfile;

void do_extract (struct fitsdata *fits, struct trace *trc, int usetrace, struct arguments *args)
{
	double *g = NULL;
	double *wav,*fdat;
	double osum;
	struct fitsdata *fwav;
	char *fln;
	struct specdata *spec;
	int i,j;
	int ord,ostart,oend,nord;
	int status = 0;
	int nrow;
	FILE *fp = NULL;
	int pwave = 0;
	int debug=0;

	fln=(char*)malloc(sizeof(char)*100);
	/*Check if a WAVE HDU exists*/
	if ((args->type <= 2) || (args->type > 3)) pwave=fitsdata_pwavexist(args->input_file);
	if (debug) {
	  if (pwave) printf("WAVE HDU exists in FITS\n");
	  else printf("_NO_ WAVE HDU in FITS\n");
	}
	/*If there is a wave array, then extract that information*/
	if (args->type==1) args->cenrow=(args->maxrow-args->minrow)/2.+args->minrow;
	wav=(double *)calloc(fits->cols,sizeof(double));
	if (pwave) {
	  fwav=fitsdata_readwave(args->input_file);
	  fdat= fwav->data;
	}
	
	if (debug) {
	  fprintf(stderr,"Input:  extype=%i   cenrow=%g  platw=%g  basew=%g\n",
		 args->type,args->cenrow,args->platw,args->basew);
	  fprintf(stderr,"        minrow=%i  maxrow=%i\n",args->minrow,args->maxrow);
	}
	if (args->type > 3) {

	  /*Echelle Extractions*/
	  nord=fits->rows/fits->ordsep;
	  for (ord=nord;ord>0;ord--) {/* Do each order one at a time */
	    sprintf(fln,"%s.%i",args->output_file,nord-ord+1);
	    /* This is setup to step backwards so that the order at the top of the CCD is #1 */
	    ostart=(ord-1)*fits->ordsep;
	    oend=(ord*fits->ordsep)-1;
	    if (debug) printf("Extracting order %i from %i to %i\n",nord-ord+1,ostart,oend);
	    if (args->type == 5) {
	      /* Mesa echelle extract */
              g = fitsdata_calcg (fits, args->cenrow+ostart, args->platw, args->basew);
	      for (j=0;j<ostart;j++) g[j]=0;
	      for (j=oend+1;j<fits->rows;j++) g[j]=0;
	      spec = fitsdata_extract (fits, g);
	    } else {
	      /* Square echelle extract */
	      g = (double *) calloc (fits->rows, sizeof (double));
	      for (i = args->minrow+ostart; i <= args->maxrow+ostart; i++)
		g[i] = 1.0;
	      spec = fitsdata_extract (fits, g);
	    }	   
 
	    free(g);

	    osum=0;
	    for (j=0;j<fits->cols;j++) 
	      if (!isnan(spec->data[j])) osum+=spec->data[j];
	    if (pwave) {
	      nrow=fits->cols*(ostart+(int)(fits->ordsep/2));
	      if (debug) printf("Getting wav from row %i\n",(ostart+(int)(fits->ordsep/2)));
	      if (fwav->rows>1) for (i=0;i<fits->cols;i++) wav[i]=fdat[i+nrow];
	      else for (i=0;i<fits->cols;i++) wav[i]=fdat[i];
	    }

	    if (osum==0) {
	      fprintf (stderr, "ERROR: Order `%i' contains no data....skipping!\n", nord-ord+1);
	    } else {
	      if (args->fits == 1)
		{
		  if (args->output_file == NULL)
		    {
		      status = specdata_write_fits( spec, "-" );
		    }
		  else
		    {
		      sprintf(fln,"%s.%i.fits",args->output_file,nord-ord+1);	  
		      status = specdata_write_fits( spec, fln );
		    }
		}
	      else if (args->unifspec == 1)
		{
		  /* Abstract out to fitstools? */
		  struct unifile u;
		  
		  u.x1 = spec->val0;
		  u.dx = spec->dval;
		  u.num_elems = spec->length;
		  u.header = NULL;
		  u.header_lines = 0;
		  u.data = spec->data;
		  /* Prepare the Unifile data structures */
		  uni_calc_meta( &u );
		  
		  if (args->output_file == NULL)
		    {
		      status = uni_write_file("-", &u);
		    }
		  else
		    {
		      sprintf(fln,"%s.%i.uni",args->output_file,nord-ord+1);	  
		      status = uni_write_file(fln, &u);
		    }
		  
		  return;
		}
	      else
		{
		  /* text file */
		  if (args->output_file == NULL)
		    {
		      fp = stdout;
		  }
		  else
		    {
		      fp = fopen (fln, "r");
		      if (fp)
			{
			  fprintf (stderr, "ERROR: Output file `%s' already exists.\n", fln);
			  fclose (fp);
			  return;
			}
		      fp = fopen (fln, "w");
		      if (!fp)
			{
			  fprintf (stderr, "ERROR: Unable to open output file `%s' for writing.\n", fln);
			  return;
			}
		    }
		  
		  status = write_header (fp, args,nord-ord+1,usetrace);
		  if (pwave) status=specdata_printwave (spec, wav, fp);
		  else status = specdata_print (spec, fp);
		  
		  if (fp != stdout)
		    fclose( fp );
		}
	    }
	  } /*End loop through orders*/

	} else {

	  /*All other extractions*/
	  if (pwave) {
	    nrow=fits->cols*(args->cenrow-1);
	    if (fwav->rows>1) for (i=0;i<fits->cols;i++) wav[i]=fdat[i+nrow];
	    else for (i=0;i<fits->cols;i++) wav[i]=fdat[i];
	  }
	  switch (args->type)
	    {
	    case 1:
	      /* Square row extract */
	      g = (double *) calloc (fits->rows, sizeof (double));
	      for (i = args->minrow; i <= args->maxrow; i++)
		g[i] = 1.0;
	      spec = fitsdata_extract (fits, g);
	      break;
	    case 2:
	      /* Mesa row extract */	      
              if (usetrace) {
		g = (double *)malloc(sizeof(double)*10);
                g[0]=trc->roff;
                g[1]=args->platw;
                g[2]=args->basew;
                args->cenrow=trc->roff;
		spec = fitsdata_extracttrace (fits, trc, g);
              } else {
		g = fitsdata_calcg (fits, args->cenrow, args->platw, args->basew);
		spec = fitsdata_extract (fits, g);
	      }
	      break;
	    case 3:
	      /* Square col extract */
	      g = (double *) calloc (fits->cols, sizeof (double));
	      for (i = args->minrow; i <= args->maxrow; i++)
		g[i] = 1.0;
	      spec = fitsdata_vertical_extract (fits, g);
	      break;
	    default:
	      fprintf (stderr, "ERROR: invalid type `%d'\n", args->type);
	      return;
	    }
	  
	  free (g);
	  if (args->fits == 1)
	    {
	      if (args->output_file == NULL)
		{
		  status = specdata_write_fits( spec, "-" );
		}
	      else
		{
		  status = specdata_write_fits( spec, args->output_file );
		}
	    }
	  else if (args->unifspec == 1)
	    {
	      /* Abstract out to fitstools? */
	      struct unifile u;

	      u.x1 = spec->val0;
	      u.dx = spec->dval;
	      u.num_elems = spec->length;
	      u.header = NULL;
	      u.header_lines = 0;
	      u.data = spec->data;
	      /* Prepare the Unifile data structures */
	      uni_calc_meta( &u );
	      
	      if (args->output_file == NULL)
		{
		  status = uni_write_file("-", &u);
		}
	      else
		{
		  status = uni_write_file(args->output_file, &u);
		}
	      
	      return;
	    }
	  else
	    {
	      /* text file */
	      if (args->output_file == NULL)
		{
		  fp = stdout;
		}
	      else
		{
		  fp = fopen (args->output_file, "r");
		  if (fp)
		    {
		      fprintf (stderr, "ERROR: Output file `%s' already exists.\n", args->output_file);
		      fclose (fp);
		      exit (1);
		    }
		  fp = fopen (args->output_file, "w");
		  if (!fp)
		    {
		      fprintf (stderr, "ERROR: Unable to open output file `%s' for writing.\n", args->output_file);
		      return;
		    }
		}
	      
	      status = write_header (fp, args,0, usetrace);
	      if (pwave) status=specdata_printwave (spec, wav, fp);
	      else status = specdata_print (spec, fp);
	      
	      if (fp != stdout)
		fclose( fp );
	    }
	}

	if (logfile)
		{
		  fprintf (logfile, "[%s] INFO : Extraction summary\n", logfile_time());
		  if (args->type > 3) fprintf(logfile,"[%s]  INFO : Echelle extraction (each order put in a differnt file)\n", logfile_time());
		  if (args->type == 3)
		    {
		      fprintf (logfile, "[%s] INFO : Vertical extraction.\n", logfile_time());
		    }
		  if ((args->type == 2) || (args->type == 4))
		    {
		      fprintf (logfile, "[%s] INFO : Simple extraction selected.\n", logfile_time());
		      fprintf (logfile, "[%s] Min row, max row: %d, %d\n", logfile_time(), args->minrow, args->maxrow );
		    }
		  if ((args->type == 1) || (args->type == 5))
		    {
		      if (usetrace) fprintf(logfile, "[%s] Used spectral trace with offest of %g rows from center.\n",logfile_time(),trc->roff);
		      else fprintf (logfile, "[%s] INFO : Central row for extraction : %.3lf\n", logfile_time(), args->cenrow + 1);

		      fprintf (logfile, "[%s] Base width, peak width for mesa: %.3lf, %.3lf\n", logfile_time(), args->basew, args->platw);
		    }
		  if (args->type == 6) fprintf(logfile,"[%s] INFO : All rows extracted in each order\n", logfile_time());
		}

	return;
}
	
